import java.util.*;

/**
 * Maintain the environment for a Game of Life.
 * 
 * @author David J. Barnes
 * @version 2009.04.21
 */
public class Environment
{
    private static final int DEFAULT_ROWS = 10;
    private static final int DEFAULT_COLS = 10;
    private static final double RANDOM_SET_PROBABILITY = 0.2;
    
    private int rows, cols;
    private Agent[][] agents;
    private EnvironmentView view;

    /**
     * Create an environment with the default size.
     */
    public Environment()
    {
        this(DEFAULT_ROWS, DEFAULT_COLS);
    }

    /**
     * Create an environment with the given size.
     * @param rows The number of rows.
     * @param cols The number of cols;
     */
    public Environment(int rows, int cols)
    {
        view = new EnvironmentView(this);
        setup(rows, cols);
    }
        
    /**
     * Run for the given number of steps.
     * @param steps The number of steps to run for.
     */
    public void run(int steps)
    {
        for(int i = 0; i < steps; i++) {
            step();
           try{
               Thread.sleep(1);
            }
            catch(InterruptedException e) {
            }
        }
    }
    
    /**
     * Run the game for one step.
     */
    public void step()
    {
        // Build a record on the next state of each agent.
        boolean[][] nextStates = new boolean[1 + rows + 1][1 + cols + 1];
        // Ask each agent to determine its next state.
        for(int row = 1; row <= rows; row++) {
            for(int col = 1; col <= cols; col++) {
                nextStates[row][col] = agents[row][col].getNextState();
            }
        }
        // Update the agents' states.
        for(int row = 1; row <= rows; row++) {
            for(int col = 1; col <= cols; col++) {
                setAgentState(row, col, nextStates[row][col]);
            }
        }
    }
    
    /**
     * Show the current state of the environment.
     */
    public void show()
    {
        for(int row = 1; row <= rows; row++) {
            for(int col = 1; col <= cols; col++) {
                System.out.print(agents[row][col].getState() ? '1' : ' ');
            }
            System.out.println();
        }
        System.out.println();
    }
    
    /**
     * Reset the state of the game to all 0.
     */
    public void reset()
    {
        for(int row = 1; row <= rows; row++) {
            for(int col = 1; col <= cols; col++) {
                setAgentState(row, col, false);
            }
        }
    }
    
    /**
     * Generate a random setup using
     * RANDOM_SET_PROBABILITY.
     */
    public void randomize()
    {
        for(int row = 1; row <= rows; row++) {
            for(int col = 1; col <= cols; col++) {
                setAgentState(row, col, Math.random() < RANDOM_SET_PROBABILITY);
            }
        }
    }
    
    /**
     * Set the state of one agent.
     * @param row The agent's row.
     * @param col The agent's col.
     * @param state The agent's state.
     */
    public void setAgentState(int row, int col, boolean state)
    {
        if(row >= 1 && row <= rows && col >= 1 && col <= cols) {
            agents[row][col].setState(state);
            view.setAgentState(row, col, state);
        }
    }
    
    /**
     * Toggle the state of one agent.
     * @param row The agent's row.
     * @param col The agent's col.
     */
    public void toggleAgentState(int row, int col)
    {
        if(row >= 1 && row <= rows && col >= 1 && col <= cols) {
            setAgentState(row, col, !agents[row][col].getState());
        }
    }
    
    /**
     * Setup a new environment of the given size.
     * @param rows The number of rows.
     * @param cols The number of cols;
     */
    private void setup(int rows, int cols)
    {
        this.rows = rows;
        this.cols = cols;
        agents = new Agent[1 + rows + 1][1 + cols + 1];
        for(int row = 0; row < agents.length; row++) {
            for(int col = 0; col < agents[row].length; col++) {
                agents[row][col] = new Agent();
            }
        }
        setupNeighbours();
        view.setup(rows, cols);
    }
    
    /**
     * Give a list of its neighbours to each agent.
     */
    private void setupNeighbours()
    {
        for(int row = 1; row <= rows; row++) {
            for(int col = 1; col <= cols; col++) {
                // The neighbours should not include the agent at (row,col).
                List<Agent> neighbours = new ArrayList<Agent>(8);
                for(int dr = -1; dr <= 1; dr++) {
                    for(int dc = -1; dc <= 1; dc++) {
                        neighbours.add(agents[row + dr][col + dc]);
                    }
                }
                // Remove agent[row][col]
                neighbours.remove(agents[row][col]);
                agents[row][col].setNeighbours(neighbours);
            }
        }
    }

}
