import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;

/**
 * A GUI for the state of the environment, with
 * runtime controls.
 * 
 * @author David J. Barnes
 * @version 2009.04.21
 */
public class EnvironmentView extends JFrame
{
    // The longest delay for the animation, in milliseconds.
    private static final int LONGEST_DELAY = 3000;
    
    private AgentView[][] agents;
    private Environment env;
    private boolean running;
    private int delay;
    
    /**
     * Constructor for objects of class EnvironmentView
     */
    public EnvironmentView(Environment env)
    {
        super("Conway's Game of Life");
        setDefaultCloseOperation(EXIT_ON_CLOSE);
        this.env = env;
        this.running = false;
        setDelay(50);
        setupControls();
        setVisible(false);
    }
    
    private void setupControls()
    {

        // Continuous stepping.
        final JButton run = new JButton("Run");
        run.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
            {
                if(!running) {
                    running = true;
                    try {
                        new Runner().execute();
                    }
                    catch(Exception ex) {
                    }
                }
            }
        });
        
        // Single stepping.
        final JButton step = new JButton("Step");
        step.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
            {
                running = false;
                env.step();
            }
        });
        
        // Resetting.
        final JButton reset = new JButton("Reset");
        reset.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
            {
                running = false;
                env.reset();
            }
        });
        
        // Randomize.
        final JButton randomize = new JButton("Randomize");
        randomize.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
            {
                running = false;
                env.randomize();
            }
        });
        
        Container contents = getContentPane();
        
        // A speed controller.
        final JSlider speedSlider = new JSlider(0, 100);
        speedSlider.addChangeListener(
            new ChangeListener() {
                public void stateChanged(ChangeEvent e) 
                {
                    setDelay((int) speedSlider.getValue());
                }
            }
        );
        Container speedPanel = new JPanel();
        speedPanel.setLayout(new GridLayout(2, 1));
        speedPanel.add(new JLabel("Animation Speed", SwingConstants.CENTER));
        speedPanel.add(speedSlider);
        contents.add(speedPanel, BorderLayout.NORTH);
        
        // Place the button controls.
        JPanel controls = new JPanel();
        controls.add(run);
        controls.add(step);
        controls.add(reset);
        controls.add(randomize);
        
        contents.add(controls, BorderLayout.SOUTH);
    }

    /**
     * Setup a new environment of the given size.
     * @param rows The number of rows.
     * @param cols The number of cols;
     */
    public void setup(int rows, int cols)
    {
        Container contents = getContentPane();
        JPanel agentPanel = new JPanel();
        agentPanel.setLayout(new GridLayout(rows, cols));
        agents = new AgentView[rows][cols];
        
        for(int row = 0; row < rows; row++) {
            for(int col = 0; col < cols; col++) {
                final int envRow = row + 1;
                final int envCol = col + 1;
                agents[row][col] = new AgentView();
                agentPanel.add(agents[row][col]);
                agents[row][col].addActionListener(
                    new ActionListener() {
                        public void actionPerformed(ActionEvent e)
                        {
                            running = false;
                            env.toggleAgentState(envRow, envCol);
                        }
                });
            }
        }
        contents.add(agentPanel, BorderLayout.CENTER);
        setSize(600,600);
        setVisible(true);
    }

    /**
     * Set the state of one agent.
     * @param row The agent's row.
     * @param col The agent's col.
     * @param state The agent's state.
     */
    public void setAgentState(int row, int col, boolean state)
    {
        agents[row - 1][col - 1].setState(state);
    }
    
    /**
     * Set the animation delay.
     * @param speedPercentage (100-speedPercentage) as a percentage of the LONGEST_DELAY.
     */
    private void setDelay(int speedPercentage)
    {
        delay = (int) ((100.0 - speedPercentage) * LONGEST_DELAY / 100);
    }
    
    /**
     * Provide stepping of the animation.
     */
    private class Runner extends SwingWorker<Boolean, Void>
    {
        public Boolean doInBackground()
        {
            while(running) {
                env.step();
                try {
                    Thread.sleep(delay);
                }
                catch(InterruptedException e) {
                }
            }
            return true;
        }
    }

}
