package simulation;

import java.util.ArrayList;
import java.util.List;

/**
 * Model The Gibson-Bruck process for simulating coupled chemical reactions. 
 * 
 * A number of possible models have been defined in this class.
 * The particular model to be run is set up in the run method; e.g.,
 *     initIsomerization();
 *
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class Model 
{
	// The set of reactions to be modeled. 
	private ReactionSet system;
	// The stop time of the simulation.
	private double stopTime;
	// The output sampling rate.
	private int samplingRate;
	// Plot list for molecules of interest during output.
	private List<Molecule> plotList;

	/**
	 * Set up the model with given stop time and sampling rate for output.
	 * @param maximumSteps
	 * @param samplingRate Report every samplingRate steps.
	 */
	public Model(double stopTime, int samplingRate)
	{
		assert stopTime > 0 && samplingRate > 0;
		this.stopTime = stopTime;
		this.samplingRate = samplingRate;
		this.plotList = new ArrayList<>();
		this.system = new ReactionSet();
	}

	/**
	 * Run the model until the stop time.
	 */
	public void run()
	{
		double time = 0;
		int step = 0;
		
		// Ensure the reaction set is clear.
		system.clear();
		
		// Initialise the required model.
		initIsomerization();	
		
		// Initialise the system according to the model.
		system.init();
		
		// Force garbage collection to reduce the risk of it
		// interfering with the runtime of the system.
		System.gc();
		
		// Show the initial status.
		showStatus(time);
		
		long startTime = System.currentTimeMillis();
		while(time < stopTime) {			
			Reaction r = system.findNextReaction();
			double nextReactionTime = r.getNextTime();
			time = nextReactionTime;
			if(time <= stopTime) {
				assert time != Double.POSITIVE_INFINITY : "No further reactions possible";
				r.react();
				step++;
				if(step % samplingRate == 0) {
					showStatus(time);
				}
				system.update(r, time);
			}
		}
		long endTime = System.currentTimeMillis();
		System.out.println("# Execution time: " + ((endTime - startTime) / 1000.0) + "s");
	}

	/**
	 * Show the current status based on the plot list.
	 * @param time The current time of the model.
	 */
	public void showStatus(double time)
	{
		System.out.print(time + " ");
		for(Molecule m : plotList) { 
			System.out.print(m.getNumber() + " ");
		}
		System.out.println();
	}
	
	/**
	 * Below is a set of init-methods to set up possible reactions to be modeled.
	 * Choose just one of these.
	 */
	
	/**
	 * Set up a basic isomerization.
	 *     X -> Y
	 */
	private void initIsomerization()
	{
		Molecule x = new Molecule("X", 1000);
		Molecule y = new Molecule("Y");
		
		system.addMolecule(x);
		system.addMolecule(y);
		
		plotList.add(x);
		plotList.add(y);
		
		Reaction r;
		// X -> Y
		r = system.addReaction(new Reaction(0.5));
		r.addReactant(x);
		r.addProduct(y);
	}
	
	/**
	 * Set up an example from 
	 * Gibson, M., Bruck, J.: Efficient exact stochastic simulation of chemical systems
	 * with many species and many channels.
	 * Journal of Physical Chemistry 104, 1876-1889 (2000)
	 *     A + B -> C
	 *     B + C -> D
	 *     D + E -> E + F
	 *     F -> D + G
	 *     E + G -> A
	 */
	private void initGB()
	{
		Molecule a = new Molecule("A", 1);
		Molecule b = new Molecule("B", 1);
		Molecule c = new Molecule("C", 1);
		Molecule d = new Molecule("D", 1);
		Molecule e = new Molecule("E", 1);
		Molecule f = new Molecule("F", 1);
		Molecule g = new Molecule("G", 1);

		plotList.add(a);
		plotList.add(b);
		plotList.add(c);
		plotList.add(d);
		plotList.add(e);
		plotList.add(f);
		plotList.add(g);
						
		Reaction r;
		
		// A + B -> C
		r = system.addReaction(new Reaction(1));
		r.addReactant(a);
		r.addReactant(b);
		r.addProduct(c);
		
		// B + C -> D
		r = system.addReaction(new Reaction(1));
		r.addReactant(b);
		r.addReactant(c);
		r.addProduct(d);
		
		// D + E -> E + F
		r = system.addReaction(new Reaction(1));
		r.addReactant(d);
		r.addReactant(e);
		r.addProduct(e);
		r.addProduct(f);
		
		// F -> D + G
		r = system.addReaction(new Reaction(1));
		r.addReactant(f);
		r.addProduct(d);
		r.addProduct(g);
		
		// E + G -> A
		r = system.addReaction(new Reaction(1));
		r.addReactant(e);
		r.addReactant(g);
		r.addProduct(a);
				
	}
	
	/**
	 * Set up the Lotka-Volterra reactions in the form given
	 * in Gillespie's paper:
	 *     X-bar + Y1 -> 2Y1
	 *     Y1 + Y2 -> 2Y2
	 *     Y2 -> Z
	 * Gillespie, D.T.: Exact stochastic simulation of coupled chemical reactions.
	 * Journal of Physical Chemistry 81(25), 2340-2361 (1977)
	 * 
	 */
	private void initLotka()
	{
		Molecule x = new Molecule("X", 3000, true);
		Molecule y1 = new Molecule("Y1", 1000);
		Molecule y2 = new Molecule("Y2", 1000);
		Molecule z = new Molecule("Z");
		
		system.addMolecule(x);
		system.addMolecule(y1);
		system.addMolecule(y2);
		system.addMolecule(z);
		
		plotList.add(y1);
		plotList.add(y2);
			
		Reaction r;
		
		// X-bar + Y1 -> 2Y1
		r = system.addReaction(new Reaction(10.0 / x.getNumber()));
		r.addReactant(x);
		r.addReactant(y1);
		r.addProduct(y1, 2);
		
		// Y1 + Y2 -> 2Y2
		r = system.addReaction(new Reaction(0.01));
		r.addReactant(y1);
		r.addReactant(y2);
		r.addProduct(y2, 2);
		
		// Y2 -> Z
		r = system.addReaction(new Reaction(10));
		r.addReactant(y2);
		r.addProduct(z);
	}
	
	/**
	 * Set up the Lotka-Volterra reactions in a slightly different form
	 *     prey-bar -> prey (Reproduction)
	 *     prey + predator-bar -> (Prey die through predation)
	 *     predator -> (Predator death)
	 *     prey-bar + predator-bar -> predator  (Reproduction in the presence of prey)
	 * 
	 */
	private void initLotkaVariation()
	{
		Molecule prey = new Molecule("prey", 1000);
		Molecule predator = new Molecule("predator", 500);
		
		system.addMolecule(prey);
		system.addMolecule(predator);
		
		plotList.add(prey);
		plotList.add(predator);
			
		Reaction r;
		
		// prey -> 2prey (Reproduction)
		r = system.addReaction(new Reaction(0.5));
		r.addReactant(prey);
		r.addProduct(prey, 2);
		
		// prey + predator -> predator (Prey die through predation)
		r = system.addReaction(new Reaction(0.0005));
		r.addReactant(prey);
		r.addReactant(predator);
		r.addProduct(predator);
				
		// predator -> (Predator death)
		r = system.addReaction(new Reaction(0.5));
		r.addReactant(predator);

		// prey + predator -> prey + 2predator  (Reproduction in the presence of prey)
		r = system.addReaction(new Reaction(0.00025));
		r.addReactant(prey);
		r.addReactant(predator);
		r.addProduct(prey);
		r.addProduct(predator, 2);
		
	}
	
	/**
	 * Initialise a Lotka-Volterra model with the given number
	 * of predator species in addition to one pure prey.
	 * @param numPredators The number of predators (positive).
	 */
	private void initLotkaMultispecies(int numPredators)
	{
		assert numPredators > 0;
		Molecule prey = new Molecule("prey", 1000);
		system.addMolecule(prey);
		plotList.add(prey);
		
		List<Molecule> predators = new ArrayList<>(numPredators);
		int initPop = 1000;
		for(int p = 0; p < numPredators; p++) {
			Molecule predator = new Molecule("predator" + p, initPop);		
			predators.add(predator);
			system.addMolecule(predator);
			plotList.add(predator);
		}
					
		Reaction r;
		
		// prey -> 2prey (Reproduction)
		r = system.addReaction(new Reaction(0.5));
		r.addReactant(prey);
		r.addProduct(prey, 2);

		for(Molecule predator : predators) {			
			// prey + predator -> predator (Prey die through predation)
			r = system.addReaction(new Reaction(Math.random() * 0.001));
			r.addReactant(prey);
			r.addReactant(predator);
			r.addProduct(predator);

			// prey-bar + predator-bar -> prey + 2predator  (Prey reproduction in the presence of prey)
			r = system.addReaction(new Reaction(Math.random() * 0.001));
			r.addReactant(prey);
			r.addReactant(predator);
			r.addProduct(prey);
			r.addProduct(predator, 2);
			
			// This predator becomes prey for the next predator.
			prey = predator;
		}	
		// predatorN -> (Predator death)
		r = system.addReaction(new Reaction(0.5));
		r.addReactant(predators.get(predators.size() - 1));
	}
	
}
