package simulation;

import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;

/**
 * Model Gillespie's Direct Method for simulating coupled chemical reactions. 
 * Delayed products are supported.
 * @author djb
 *
 */
public class Model
{
	// The set of reactions to be modeled. 
	private ReactionSet system;
	// The current simulation time;
	private double time;
	// The model stop time.
	private double stopTime;
	// A queue of pending products, ordered by time of production.
	private Queue<PendingProduct> delayedProducts;
	
	// The maximum number of simulation steps required.
	private int samplingRate;
	// Plot list for molecules of interest during output.
	private List<Molecule> plotList;

	/**
	 * Create a model, with stop time and sampling rate for output.
	 * @param stopTime Stop when this time is reached.
	 * @param samplingRate Output every samplingRate steps.
	 */
	public Model(double stopTime, int samplingRate)
	{
		// The set of reactions to be modeled. 
		this.system = new ReactionSet();
		this.time = 0;
		this.stopTime = stopTime;
		this.delayedProducts = new PriorityQueue<PendingProduct>();
		this.samplingRate = samplingRate;
		this.plotList = new ArrayList<Molecule>();
	}

	/**
	 * Run until the stop time, or no reaction is possible.
	 */
	public void run()
	{
		plotList.clear();
		system.clear();
				
		// Initialize the required model.
		initDelayedIsomerization();	
				
		// Calculate the initial a-zero value.
		double aZero = system.calculateAZero();
		int step = 0;
		time = 0;
		
		// Show the initial status.
		showStatus();
		
		// Record the starting time.
		long startTime = System.currentTimeMillis();

		while(time < stopTime && !(aZero == 0.0 && delayedProducts.size() == 0)) {		
			// Calculate the time to the next reaction.
			double deltaT = (1.0 / aZero * Math.log(1.0 / Math.random()));
			// See if a delayed product should be produced first.
			PendingProduct p = delayedProducts.peek();
			if(p != null && p.getProductionTime() <= time + deltaT) {
				delayedProducts.remove();
				p.produce();
				time = p.getProductionTime();
			}
			else {
				Reaction r = system.findReaction(Math.random() * aZero);
				time += deltaT;
				r.react();
			}
			step++;
			if(step % samplingRate == 0) {
				showStatus();
			}
			aZero = system.calculateAZero();
		}
		long endTime = System.currentTimeMillis();
		System.out.println("# Execution time: " + ((endTime - startTime) / 1000.0) + "s" +
						   " Steps: " + step);
	}
	
	/**
	 * Return the current simulation time.
	 */
	public double getTime()
	{
		return time;
	}
	
	/**
	 * Add the pending product to the queue for delayed production.
	 * @param product The delayed product.
	 */
	public void scheduleDelayedProduction(PendingProduct product)
	{
		delayedProducts.add(product);
	}
	
	/**
	 * Show the current status based on the plot list.
	 */
	public void showStatus()
	{
		System.out.print(time + " ");
		for(Molecule m : plotList) { 
			System.out.print(m.getNumber() + " ");
		}
		System.out.println();
	}
	
	/**
	 * Below is a set of init-methods to set up possible reactions to be modeled.
	 * Choose just one of these.
	 */
	
	/**
	 * Set up a basic isomerization.
	 *     X -> Y
	 */
	private void initIsomerization()
	{
		Molecule x = new Molecule("X", 1000);
		Molecule y = new Molecule("Y");
		
		// Show the levels of x and y in the output.
		plotList.add(x);
		plotList.add(y);
		
		// X -> Y
		Reaction r = new Reaction(0.5);
		r.addReactant(x);
		r.addProduct(y);
		system.addReaction(r);
	}
	
	/**
	 * Set up a basic isomerization alongside one with a delay.
	 *     X1 -> Y1 (immediate)
	 *     X2 -> Y2 (delayed)
	 * @param system The reaction set to be initialized.
	 */
	private void initDelayedIsomerization()
	{
		Molecule x1 = new Molecule("X1", 1000);
		Molecule y1 = new Molecule("Y1");
		Molecule x2 = new Molecule("X2", 1000);
		Molecule y2 = new Molecule("Y2");
		
		// Show the levels of x and y in the output.
		plotList.add(x1);
		plotList.add(y1);
		plotList.add(x2);
		plotList.add(y2);
		
		// X1 -> Y1
		Reaction r = new Reaction(0.5);
		r.addReactant(x1);
		r.addProduct(y1);
		system.addReaction(r);
		
		// X2 -> Y2
		r = new Reaction(0.5);
		r.addReactant(x2);
		r.addProduct(new DelayedProduct(y2, 1, 1.0, this));
		system.addReaction(r);
	}
}
