package simulation;

import java.util.ArrayList;
import java.util.List;

/**
 * Model a reaction as a combination of reactants yielding products.
 * @author djb
 *
 */
public class Reaction 
{
	// The reaction's reactants.
	private List<Reactant> reactants;
	// The reaction's products.
	private List<Product> products;
	// Stochastic reaction constant.
	private double reactionConstant;
	// The reaction's propensity
	//     = combinations * reactionConstant;
	private double propensity;
	
	/**
	 * Create a reaction with the given reaction constant.
	 * 
	 * @param c The reaction constant.
	 */
	public Reaction(double c)
	{
		this.reactants = new ArrayList<Reactant>();
		this.products = new ArrayList<Product>();
		this.reactionConstant = c;
	}
	
	/**
	 * Add a molecule as a reactant, with multiplicity of 1.
	 * @param m The reactant molecule.
	 */
	public void addReactant(Molecule m)
	{
		addReactant(m, 1);
	}
	
	/**
	 * Add a molecule as a reactant, with given multiplicity.
	 * @param m The reactant molecule.
	 * @param multiplicity
	 */
	public void addReactant(Molecule m, int multiplicity)
	{
		reactants.add(new Reactant(m, multiplicity));
	}
	
	/**
	 * Add a molecule as a product, with multiplicity of 1.
	 * @param m The reactant molecule.
	 */
	public void addProduct(Molecule m)
	{
		addProduct(m, 1);
	}
	
	/**
	 * Add a molecule as a product, with given multiplicity.
	 * @param m The reactant molecule.
	 * @param multiplicity
	 */
	public void addProduct(Molecule m, int multiplicity)
	{
		addProduct(new Product(m, multiplicity));
	}
	
	/**
	 * Add the given product to the reaction.
	 * @param p The product to be added.
	 */
	public void addProduct(Product p)
	{
		products.add(p);
	}
	
	/**
	 * Return the current propensity.
	 * @return The current propensity.
	 */
	public double getPropensity()
	{
		return propensity;
	}
	
	/**
	 * Calculate and set the value of the propensity.
	 * @return The new value of the propensity.
	 */
	public double calculatePropensity()
	{
		return propensity = calculateCombinations() * reactionConstant;
	}
	
	/**
	 * Apply this reaction, as long as there are sufficient reactants.
	 */
	public void react()
	{
		// Check that there are sufficient reactants, and die if not.
		for(Reactant r : reactants) {
			if(r.getMultiplicity() > r.getMolecule().getNumber()) {
				throw new RuntimeException("Impossible reaction selected: " + this);
			}
		}
		
		// Consume the reactants.
		for(Reactant r : reactants) {
			r.consume();
		}
		// Produce the products.
		for(Product p : products) {
			p.produce();
		}
	}
	
	/**
	 * Calculate the number of combinations based on the numbers of participating and
	 * available reactants.
	 * @return The number of combinations.
	 */
	private double calculateCombinations()
	{
		double h = 1.0;
		for(Reactant r : reactants) {
			int m = r.getMultiplicity();
			if(m > 1) {
			    // Avoid factorial calculations.
			    int numerator = 1;
			    int denominator = 1;
			    for(int times = 0; times < m; times++) {
				    numerator *= r.getMolecule().getNumber() - times;
				    denominator *= m - times;
			    }
			    h *= numerator / denominator;
			}
			else {
				h *= r.getMolecule().getNumber();
			}
		}
		return h;
	}
	
}
