package simulation;

import java.util.ArrayList;
import java.util.List;

/**
 * Model Gillespie's Direct Method for simulating coupled chemical reactions. 
 * 
 * A number of possible models have been defined in this class.
 * The particular model to be run is set up in the run method; e.g.,
 *     initIsomerization();
 *     
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class Model
{
	// The set of reactions to be modeled. 
	private ReactionSet system;
	// The model time.
	private double time;
	// The model stop time.
	private double stopTime;
	// The output sampling rate.
	// Output the reaction status every samplingRate steps.
	private int samplingRate;
	// Plot list for molecules of interest during output.
	private List<Molecule> plotList;

	/**
	 * Create a model, with stop time and sampling rate for output.
	 * @param stopTime Stop when this time is reached.
	 * @param samplingRate Output every samplingRate steps.
	 */
	public Model(double stopTime, int samplingRate)
	{
		// The set of reactions to be modeled. 
		this.system = new ReactionSet();
		this.stopTime = stopTime;
		this.samplingRate = samplingRate;
		this.plotList = new ArrayList<Molecule>();
		this.time = 0;
	}

	/**
	 * Run until the stop time, or no reaction is possible.
	 */
	public void run()
	{
		plotList.clear();
		system.clear();
		
		// Initialize the required model.
		initIsomerization();	
				
		// Calculate the initial a-zero value.
		double aZero = system.calculateAZero();
		int step = 0;
		time = 0;
		// Show the initial status.
		showStatus(time);
		// Record the starting time.
		long startTime = System.currentTimeMillis();
		
		while(time < stopTime && aZero != 0.0) {		
			// Calculate the time to the next reaction.
			double deltaT = (1.0 / aZero * Math.log(1.0 / Math.random()));
			time += deltaT;
			Reaction r = system.findReaction(Math.random() * aZero);
			r.react();
			step++;
			if(step % samplingRate == 0) {
				showStatus(time);
			}
			aZero = system.calculateAZero();
		}
		long endTime = System.currentTimeMillis();
		System.out.println("# Execution time: " + ((endTime - startTime) / 1000.0) + "s" +
						   " Steps: " + step);
	}
	
	/**
	 * Show the current status based on the plot list.
	 * @param time The current time of the model.
	 */
	public void showStatus(double time)
	{
		System.out.print(time + " ");
		for(Molecule m : plotList) { 
			System.out.print(m.getNumber() + " ");
		}
		System.out.println();
	}
	
	/**
	 * Below is a set of init-methods to set up possible reactions to be modeled.
	 * Choose just one of these.
	 */
	
	/**
	 * Set up a basic isomerization.
	 *     X -> Y
	 */
	private void initIsomerization()
	{
		Molecule x = new Molecule("X", 1000);
		Molecule y = new Molecule("Y");
		
		// Show the levels of x and y in the output.
		plotList.add(x);
		plotList.add(y);
		
		// X -> Y
		Reaction r = new Reaction(0.5);
		r.addReactant(x);
		r.addProduct(y);
		system.addReaction(r);
	}
	
	/**
	 * Set up the Lotka reactions:
	 *     X-bar + Y1 -> 2Y1
	 *     Y1 + Y2 -> 2Y2
	 *     Y2 -> Z
	 * Gillespie, D.T.: Exact stochastic simulation of coupled chemical reactions.
	 * Journal of Physical Chemistry 81(25), 2340-2361 (1977)
	 */
	private void initLotka()
	{
		Molecule x = new Molecule("X", 3000, true);
		Molecule y1 = new Molecule("Y1", 1000);
		Molecule y2 = new Molecule("Y2", 1000);
		Molecule z = new Molecule("Z");
		
		plotList.add(x);
		plotList.add(y1);
		plotList.add(y2);
		plotList.add(z);
			
		Reaction r;
		
		// X-bar + Y1 -> 2Y1
		r = system.addReaction(new Reaction(10.0 / x.getNumber()));
		r.addReactant(x);
		r.addReactant(y1);
		r.addProduct(y1, 2);
		
		// Y1 + Y2 -> 2Y2
		r = system.addReaction(new Reaction(0.01));
		r.addReactant(y1);
		r.addReactant(y2);
		r.addProduct(y2, 2);
		
		// Y2 -> Z
		r = system.addReaction(new Reaction(10));
		r.addReactant(y2);
		r.addProduct(z);
	}
	
	/**
	 * Set up a miscellaneous, arbitrary set of reactions:
	 * 		A + B-bar -> C
	 * 		A + D -> E
	 * 		C + E -> 2A + D
	 */
	private void initMisc()
	{
		Molecule a = new Molecule("A", 1000);
		Molecule b = new Molecule("B", 1000, true);
		Molecule c = new Molecule("C", 1000);
		Molecule d = new Molecule("D", 1000);
		Molecule e = new Molecule("E", 1000);
		
		plotList.add(a);
		plotList.add(c);
		plotList.add(d);
		plotList.add(e);
		
		Reaction r;
		
		// A + B-bar -> C
		r = system.addReaction(new Reaction(0.12));
		r.addReactant(a);
		r.addReactant(b);
		r.addProduct(c);
		
		// A + D -> E
		r = system.addReaction(new Reaction(0.08));
		r.addReactant(a);
		r.addReactant(d);
		r.addProduct(e);
		
		// C + E -> 2A + D
		r = system.addReaction(new Reaction(0.5));
		r.addReactant(c);
		r.addReactant(e);
		r.addProduct(a, 2);
		r.addProduct(d);		
	}
}
