package simulation;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * Model The SSA-CR process for simulating coupled chemical reactions.
 * Described in Slepoy et al, J. Chem. Phys. 128, 205101 (2008) 
 * 
 * A number of possible models have been defined in this class.
 * The particular model to be run is set up in the run method; e.g.,
 *     initIsomerization();
 *
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class Model 
{
	// The set of reactions to be modeled. 
	private ReactionSet system;
	// The simulation stop time.
	private double stopTime;
	// The output sampling rate.
	private int samplingRate;
	// Plot list for molecules of interest during output.
	private List<Molecule> plotList;

	/**
	 * Set up the model with given stop time and sampling rate for output.
	 * @param stopTime The simulation stop time.
	 * @param samplingRate Report every samplingRate steps.
	 */
	public Model(double stopTime, int samplingRate)
	{
		assert stopTime > 0 && samplingRate > 0;
		this.stopTime = stopTime;
		this.samplingRate = samplingRate;
		this.plotList = new ArrayList<Molecule>();
		this.system = new ReactionSet();
	}

	/**
	 * Run the model until the stop time.
	 */
	public void run()
	{
		double time = 0;
		int step = 0;
		
		// Initialise the required model.
		initIsomerization();	
		
		// Initialise the set of reactions.
		system.initReactionSet();
		
		// Force garbage collection to reduce the risk of it
		// interfering with the runtime of the system.
		System.gc();
		
		// Show the initial status.
		showStatus(time);
		
		long startTime = System.currentTimeMillis();
		while(time < stopTime) {			
			double deltaT = system.getDeltaT();
			assert deltaT > 0;
			Reaction r = system.findNextReaction();
			time += deltaT;
			r.react();
			system.update(r);
			step++;
			if(step % samplingRate == 0) {
				showStatus(time);
			}
		}
		long endTime = System.currentTimeMillis();
		System.out.println("# Execution time: " + ((endTime - startTime) / 1000.0) + "s");
	}

	/**
	 * Show the current status based on the plot list.
	 * @param time The current time of the model.
	 */
	public void showStatus(double time)
	{
		System.out.print(time + " ");
		for(Molecule m : plotList) { 
			System.out.print(m.getNumber() + " ");
		}
		System.out.println();
	}
	
	/**
	 * Below is a set of init-methods to set up possible reactions to be modeled.
	 * Choose just one of these.
	 */
	
	/**
	 * Set up a basic isomerization.
	 *     X -> Y
	 */
	private void initIsomerization()
	{
		Molecule x = new Molecule("X", 1000);
		Molecule y = new Molecule("Y");
		
		system.addMolecule(x);
		system.addMolecule(y);
		
		plotList.add(x);
		plotList.add(y);
		
		Reaction r;
		// X -> Y
		r = system.addReactionToSet(new Reaction(0.5));
		r.addReactant(x);
		r.addProduct(y);
	}
	

	/**
	 * Set up an example from 
	 * Gibson, M., Bruck, J.: Efficient exact stochastic simulation of chemical systems
	 * with many species and many channels.
	 * Journal of Physical Chemistry 104, 1876-1889 (2000)
	 *     A + B -> C
	 *     B + C -> D
	 *     D + E -> E + F
	 *     F -> D + G
	 *     E + G -> A
	 */
	private void initGB()
	{
		Molecule a = new Molecule("A", 1);
		Molecule b = new Molecule("B", 1);
		Molecule c = new Molecule("C", 1);
		Molecule d = new Molecule("D", 1);
		Molecule e = new Molecule("E", 1);
		Molecule f = new Molecule("F", 1);
		Molecule g = new Molecule("G", 1);

		plotList.add(a);
		plotList.add(b);
		plotList.add(c);
		plotList.add(d);
		plotList.add(e);
		plotList.add(f);
		plotList.add(g);
						
		Reaction r;
		
		// A + B -> C
		r = system.addReactionToSet(new Reaction(1));
		r.addReactant(a);
		r.addReactant(b);
		r.addProduct(c);
		
		// B + C -> D
		r = system.addReactionToSet(new Reaction(1));
		r.addReactant(b);
		r.addReactant(c);
		r.addProduct(d);
		
		// D + E -> E + F
		r = system.addReactionToSet(new Reaction(1));
		r.addReactant(d);
		r.addReactant(e);
		r.addProduct(e);
		r.addProduct(f);
		
		// F -> D + G
		r = system.addReactionToSet(new Reaction(1));
		r.addReactant(f);
		r.addProduct(d);
		r.addProduct(g);
		
		// E + G -> A
		r = system.addReactionToSet(new Reaction(1));
		r.addReactant(e);
		r.addReactant(g);
		r.addProduct(a);
				
	}
	

	/**
	 * Set up the Lotka reactions:
	 *     X-bar + Y1 -> 2Y1
	 *     Y1 + Y2 -> 2Y2
	 *     Y2 -> Z
	 * Gillespie, D.T.: Exact stochastic simulation of coupled chemical reactions.
	 * Journal of Physical Chemistry 81(25), 2340-2361 (1977)
	 * 
	 */
	private void initLotka()
	{
		Molecule x = new Molecule("X", 3000, true);
		Molecule y1 = new Molecule("Y1", 1000);
		Molecule y2 = new Molecule("Y2", 1000);
		Molecule z = new Molecule("Z");
		
		system.addMolecule(x);
		system.addMolecule(y1);
		system.addMolecule(y2);
		system.addMolecule(z);
		
		plotList.add(y1);
		plotList.add(y2);
			
		Reaction r;
		
		// X-bar + Y1 -> 2Y1
		r = system.addReactionToSet(new Reaction(10.0 / x.getNumber()));
		r.addReactant(x);
		r.addReactant(y1);
		r.addProduct(y1, 2);
		
		// Y1 + Y2 -> 2Y2
		r = system.addReactionToSet(new Reaction(0.01));
		r.addReactant(y1);
		r.addReactant(y2);
		r.addProduct(y2, 2);
		
		// Y2 -> Z
		r = system.addReactionToSet(new Reaction(10));
		r.addReactant(y2);
		r.addProduct(z);
	}
	
	/**
	 * Create a set of random reactions.
	 */
	private void initRandomSet()
	{
		final int NUM_RANDOM_REACTIONS = 100000;
		final int NUM_MOLECULES = NUM_RANDOM_REACTIONS / 4;
		final int MAX_MOLECULES = 500;
		final double MAX_PROPENSITY = 100.0;
		Random rand = new Random();
		
		for(int i = 0; i < NUM_MOLECULES; i++) {
			String name = "M" + i;
			Molecule m = new Molecule(name, 1 + rand.nextInt(MAX_MOLECULES));
			system.addMolecule(m);
			//plotList.add(m);
		}
		
		for(int i = 0; i < NUM_RANDOM_REACTIONS; i++) {
			Reaction r = system.addReactionToSet(new Reaction(rand.nextDouble() * MAX_PROPENSITY));
			Molecule reactant1 = system.getMolecule("M" + rand.nextInt(NUM_MOLECULES));
			Molecule reactant2 = system.getMolecule("M" + rand.nextInt(NUM_MOLECULES));

			if(reactant1 != reactant2) {
				r.addReactant(reactant1);
				r.addReactant(reactant2);
			}
			else {
				r.addReactant(reactant1, 2);
			}
			
			Molecule product1 = system.getMolecule("M" + rand.nextInt(NUM_MOLECULES));
			Molecule product2 = system.getMolecule("M" + rand.nextInt(NUM_MOLECULES));
			if(product1 != product2) {
				r.addProduct(product1);
				r.addProduct(product2);
			}
			else {
				r.addProduct(product1, 2);				
			}
		}
	}

}
