package simulation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;
import java.util.List;

/**
 * Model a reaction as a combination of reactants yielding products.
 * Either reactants or products may be empty, but not both.
 *
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class Reaction
{
	// The reaction's reactants. May be empty.
	private List<Reactant> reactants;
	// The reaction's products. May be empty.
	private List<Product> products;
	// The reaction's dependency set.
	private Set<Molecule> dependsOn;
	// The reaction's affects set.
	private Set<Molecule> affects;
	// Reaction constant.
	// For generality, it should be possible for this to vary.
	// This can be achieved by overriding the calculatePropensity method.
	private double reactionConstant;
	// propensity = combinations * reactionConstant;
	private double propensity;	
	// The last propensity.
	private double oldPropensity;
	// The id of this reaction.
	private int id;
	
	// This reaction's group.
	private PropensityGroup group;
	// The index of this reaction in its group.
	private int groupIndex;
	
	/**
	 * Set up for the reactants and products of a reaction.
	 * The reaction is not complete until the reactants and/or products
	 * have been added.
	 * @param reactionConstant The reaction constant.
	 */
	public Reaction(double reactionConstant)
	{
		this.reactants = new ArrayList<Reactant>();
		this.products = new ArrayList<Product>();
		this.dependsOn = new HashSet<Molecule>();
		this.affects = new HashSet<Molecule>();
		this.reactionConstant = reactionConstant;
		this.oldPropensity = 0;
	}
	
	/**
	 * Add a reactant with multiplicity 1.
	 * @param m The reactant.
	 */
	public void addReactant(Molecule m)
	{
		addReactant(m, 1);
	}
	
	/**
	 * Add a reactant with given multiplicity.
	 * @param m The reactant.
	 * @param multiplicity Positive multiplicity.
	 */
	public void addReactant(Molecule m, int multiplicity)
	{
		assert multiplicity > 0;
		reactants.add(new Reactant(m, multiplicity));
		if(!m.isBoundarySpecies()) {
			addToDependsOn(m);
			addToAffects(m);
		}
	}
	
	/**
	 * Add a product with multiplicity 1.
	 * @param m The product.
	 */
	public void addProduct(Molecule m)
	{
		addProduct(m, 1);
	}
	
	/**
	 * Add a product with given multiplicity.
	 * @param m The product.
	 * @param multiplicity Positive multiplicity.
	 */
	public void addProduct(Molecule m, int multiplicity)
	{
		assert multiplicity > 0;
		products.add(new Product(m, multiplicity));
		addToAffects(m);
	}
		
	/**
	 * Return the propensity.
	 * @return The propensity.
	 */
	public double getPropensity()
	{
		return propensity;
	}
	
	/**
	 * Return the old propensity from prior to the last reaction.
	 * @return The old propensity.
	 */
	public double getOldPropensity()
	{
		return oldPropensity;
	}
	
	/**
	 * Calculate and set the value of the propensity.
	 * @return The new value of the propensity.
	 */
	public double calculatePropensity()
	{
		oldPropensity = propensity;
		return propensity = calculateCombinations() * getReactionConstant();
	}
	
	/**
	 * Return the reaction constant.
	 * @return The reaction constant.
	 */
	public double getReactionConstant()
	{
		return reactionConstant;
	}

	/**
	 * Return the set of molecules this reaction depends upon.
	 * @return The set of molecules this reaction depends upon.
	 */
	public Set<Molecule> getDependsOn() 
	{
		return dependsOn;
	}

	/**
	 * Return the set of molecules this reaction affects.
	 * @return The set of molecules this reaction affects.
	 */
	public Set<Molecule> getAffects() 
	{
		return affects;
	}

	/**
	 * Apply this reaction, as long as there are sufficient reactants.
	 */
	public void react()
	{
		// Check that there are sufficient reactants.
		for(Reactant r : reactants) {
			if(r.getMultiplicity() > r.getMolecule().getNumber()) {
				throw new RuntimeException("Impossible reaction selected: " + this);
			}
		}
		
		// Consume the reactants.
		for(Reactant r : reactants) {
			if(r.isConsumed()) {
				r.consume();
			}
		}
		// Produce the products.
		for(Product p : products) {
			p.produce();
		}
	}
	
	/**
	 * Build a string representation of this reaction.
	 * @return a string representation of this reaction.
	 */
	public String toString()
	{
		StringBuilder builder = new StringBuilder();
		Iterator<Reactant> itR = reactants.iterator();
		builder.append(itR.next());
		while(itR.hasNext()) {
			builder.append(" + ");
			builder.append(itR.next());
		}
		builder.append(" -> ");
		Iterator<Product> itP = products.iterator();
		builder.append(itP.next());
		while(itP.hasNext()) {
			builder.append(" + ");
			builder.append(itP.next());
		}
		builder.append(" (");
		builder.append(getPropensity());
		builder.append(")");
		return builder.toString();
	}
	
	/**
	 * Calculate the minimum propensity.
	 * @return The minimum propensity.
	 */
	public double calculateMinimumPropensity()
	{
		return calculateMinimumCombinations() * getReactionConstant();
	}
	
	/**
	 * Return this reaction's group
	 * @return the group (may be null)
	 */
	public PropensityGroup getGroup()
	{
		return group;
	}

	/**
	 * Set this reaction's group.
	 * @param group the group to set (may be null)
	 */
	public void setGroup(PropensityGroup group)
	{
		this.group = group;
	}

	/**
	 * Get this reaction's index in its group.
	 * @return
	 */
	public int getGroupIndex()
	{
		return groupIndex;
	}

	/**
	 * Set this reaction's index in its group.
	 * @return
	 */
	public void setGroupIndex(int groupIndex)
	{
		this.groupIndex = groupIndex;
	}

	/**
	 * Add the molecule to the list of molecules this reaction depends on.
	 * @param m The molecule this reaction depends on.
	 */
	protected void addToDependsOn(Molecule m)
	{
		dependsOn.add(m);
	}
	
	/**
	 * Calculate the number of combinations based on the numbers of participating and
	 * available reactants.
	 * @return The number of combinations.
	 */
	private double calculateCombinations()
	{
		// If there are no explicit reactants then 1.0 is returned.
		double combinations = 1.0;
		for(Reactant r : reactants) {
			int m = r.getMultiplicity();
			if(m > 1) {
				// Avoid factorial calculations.
				int numerator = 1;
				int denominator = 1;
				for(int times = 0; times < m; times++) {
					numerator *= r.getMolecule().getNumber() - times;
					denominator *= m - times;
				}
				combinations *= numerator / denominator;
			}
			else {
				combinations *= r.getMolecule().getNumber();
			}
		}
		return combinations;
	}
	
	/**
	 * Calculate the minimum number of combinations based on the numbers of
	 *  participating and available reactants.
	 * @return The minimum number of combinations.
	 */
	private double calculateMinimumCombinations()
	{
		// If there are no explicit reactants then 1.0 is returned.
		double combinations = 1.0;
		// Instead of using the actual number of reactant molecules, use
		// the multiplicities to give the minimum, non-zero, number of combinations.
		for(Reactant r : reactants) {
			int mult = r.getMultiplicity();
			int number = r.isConsumed() ? mult : r.getMolecule().getNumber();
			if(number > mult) {
				if(mult > 1) {
					// Multiply by ...
					// Avoid factorial calculations.
					int numerator = 1;
					int denominator = 1;
					for(int times = 0; times < mult; times++) {
						numerator *= number - times;
						denominator *= mult - times;
					}
					combinations *= numerator / denominator;
				}
				else {
					combinations *= number;
				}
			}
			else {
				// Only one combination.
			}
		}
		return combinations;
	}
	
	/**
	 * Add the molecule to the list of molecules affected by this reaction.
	 * However, if m is catalytic in this reaction then it should not occur
	 * in affects.
	 * @param m The molecule affected.
	 */
	private void addToAffects(Molecule m)
	{
		// Look for catalytic reactions.
		boolean catalytic = false;
		Reactant reactant = null;
		Product product = null;
		for(Reactant r : reactants) {
			if(r.getMolecule() == m) {
				reactant = r;
			}
		}
		for(Product p : products) {
			if(p.getMolecule() == m) {
				product = p;
			}
		}
		if(reactant != null && product != null) {
			if(reactant.getMultiplicity() == product.getMultiplicity() &&
					reactant.isConsumed()) {
				catalytic = true;
			}
		}

		if(!catalytic){
			affects.add(m);
		}
		else {
			// Remove any previous addition.
			affects.remove(m);
		}
	}

	/**
	 * Set this reaction's id.
	 * @param id the id to set
	 */
	public void setId(int id)
	{
		this.id = id;
	}

	/**
	 * Return this reaction's id.
	 * @return the id
	 */
	public int getId()
	{
		return id;
	}

}
