package simulation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

/**
 * Maintain a set of interacting reactions.
 *
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class ReactionSet 
{
	// A convenience constant.
	private static final double LOG2 = Math.log(2);
	
	// The reactions.
	private List<Reaction> reactions;
	// A convenience map of molecule names.
	private Map<String, Molecule> molecules;

	// The dependency graph between reactions.
	private Reaction[][] dependencyGraph;
	//
	// The minimum propensity;
	private double minimumPropensity;
	// The list of groups.
	private ArrayList<PropensityGroup> groups;
	// The tree of cumulative propensity values.
	private double[] propensityTree;
	// The offset in propensityTree of the first leaf
	// -- i.e. propensityTree[propensityTreeLeafOffset]
	// is the values for groups[0].
	private int propensityTreeLeafOffset;
	// Collect the nodes whose propensity values must be updated because
	// of a change to one of their children.
	private boolean[] changeList;
	
	// Random number generator for group and reaction selection.
	private Random rand;
	
	/**
	 * Create a reaction set ready to be initialized via
	 * initReactionSet().
	 */
	public ReactionSet()
	{
		this.reactions = new ArrayList<Reaction>();
		this.molecules = new HashMap<String, Molecule>();
		this.groups = new ArrayList<PropensityGroup>();
		this.rand = new Random();
	}
		
	/**
	 * Initialize the required reaction set.
	 */
	public void initReactionSet()
	{
		// Initialize the required model.
		buildDependencyGraph();		
		
		for(Reaction r : reactions) {
			r.calculatePropensity();
		}
		
		calculateMinimumPropensity();
		
		// Allocate each reaction to a group with similar propensity.
		initGroups();
		// Calculate the accumulated propensity for all groups.
		buildPropensityTree();
	}
	
	/**
	 * Return the time delta of the next reaction.
	 * @return the time delta of the next reaction
	 */
	public double getDeltaT()
	{
		assert propensityTree[0] != 0 : "No further reactions possible";
		double deltaT = (1.0 / propensityTree[0] * Math.log(1.0 / rand.nextDouble()));
		assert deltaT > 0 : "Total propensity: " + propensityTree[0];
		return deltaT;
	}
	
	/**
	 * Find the next reaction using the composition and rejection approach. 
	 * @return The selected reaction.
	 */
	public Reaction findNextReaction()
	{
		assert propensityTree[0] != 0;
		// Generate a random propensity within the full range.
		double accumulatedPropensity = rand.nextDouble() * propensityTree[0];
		// Descend the propensity tree looking for the selected group at the leaf level.
		int node = 0;
		while(node < propensityTreeLeafOffset) {
			int leftChild = 2 * node + 1;
			double leftChildPropensity = propensityTree[leftChild];
			if(accumulatedPropensity < leftChildPropensity) {
				node = leftChild;
			}
			else {
				node = leftChild + 1;
				accumulatedPropensity -= leftChildPropensity;
			}
		}
		PropensityGroup group = groups.get(node - propensityTreeLeafOffset);
		return group.selectReaction(rand);				
	}
	
	/**
	 * Update the system according to the effect of the last
	 * reaction.
	 * @param lastReaction The last reaction.
	 */
	public void update(Reaction lastReaction)
	{
		// lastReaction will be in its own dependency graph,
		// so there is no need to treat is separately.
		for(Reaction r : dependencyGraph[lastReaction.getId()]) {
			// Calculate the dependent reaction's new propensity.
			r.calculatePropensity();
			// Update r within the group list.
			updateGroups(r);
		}
		updatePropensityTree();
	}
	
	/**
	 * Update the propensity tree following changes to
	 * group propensities as a result of the last reaction.
	 */
	private void updatePropensityTree()
	{
		// Percolate the changes to the groups' propensities up the tree.
		boolean[] shadowChangeList = new boolean[changeList.length];
		double[] shadowPropensityTree = new double[propensityTree.length];
		System.arraycopy(changeList, 0, shadowChangeList, 0, changeList.length);
		System.arraycopy(propensityTree, 0, shadowPropensityTree, 0, propensityTree.length);
		int leftChild = 2 * (propensityTreeLeafOffset - 1) + 1;
		for(int node = propensityTreeLeafOffset - 1; node > 0; node--) {
			if(shadowChangeList[node]) {
				shadowPropensityTree[node] = shadowPropensityTree[leftChild] + shadowPropensityTree[leftChild + 1];
				shadowChangeList[parentNode(node)] = true;
				shadowChangeList[node] = false;
			}
			leftChild -= 2;
		}
		// Adjust the root.
		shadowPropensityTree[0] = shadowPropensityTree[1] + shadowPropensityTree[2];
		
		for(int leaf = 0; leaf < changeList.length; leaf++) {
			if(changeList[leaf]) {
				percolateChangedNodeValue(leaf);
				changeList[leaf] = false;
			}
		}
		if(!Arrays.equals(propensityTree, shadowPropensityTree)) {
			throw new RuntimeException("Mismatch");
		}
	}
	
	/**
	 * Add a molecule to the reaction set.
	 * @param m The molecule to be added.
	 */
	public void addMolecule(Molecule m)
	{
		molecules.put(m.getName(), m);		
	}
	
	/**
	 * Get the molecule with the given name.
	 * @param name The name of the molecule
	 * @return Return the given molecule.
	 */
	public Molecule getMolecule(String name)
	{
		return molecules.get(name);
	}
	
	/**
	 * Add a new reaction to the set of reactions.
	 * @param The reaction to be added.
	 * @return The reaction added.
	 */
	public Reaction addReactionToSet(Reaction r)
	{
		// Give the reaction an ID that matches its (fixed) location
		// in the list of reactions.
		r.setId(reactions.size());
		reactions.add(r);
		return r;
	}
		
	/**
	 * Calculate the minimum propensity of the reaction set.
	 */
	private void calculateMinimumPropensity()
	{
		minimumPropensity = Double.MAX_VALUE;
		for(Reaction r : reactions) {
			double propensity = r.calculateMinimumPropensity();
			if(propensity < minimumPropensity) {
				minimumPropensity = propensity;
			}
		}
		// Place a lower-bound on the minimum.
		minimumPropensity = Math.max(1.0, minimumPropensity);
	}
	
	/**
	 * Allocate reactions to groups according to propensity.
	 */
	private void initGroups()
	{
		for(Reaction r : reactions) {
			double propensity = r.getPropensity();
			if(true || propensity > minimumPropensity) {
				int groupId = (int) logBase2(propensity / minimumPropensity);
				// Make sure the group exists.
				while(groups.size() <= groupId) {
					int groupNumber = groups.size();
					double boundary = Math.pow(2.0, groupNumber + 1) * minimumPropensity;
					groups.add(new PropensityGroup(groupNumber, boundary));
				}
				PropensityGroup group = groups.get(groupId);
				group.addReaction(r);
			}
			else {
				// Consider it unreactable.
				// NB: It could be put in group 0, but with a low probability
				// of selection.
				r.setGroup(null);
			}
		}
		// Provide enough space for the number of groups to expand to a power of 2.
		int nextPowerOf2 = (int) Math.pow(2.0, (int) Math.ceil(logBase2(groups.size())));
		groups.ensureCapacity(nextPowerOf2);
		// Create a place to record which groups require their propensity
		// values to be updated as a result of the current reaction.
		changeList = new boolean[nextPowerOf2];
	}
	
	/**
	 * The reaction has just adjusted its propensity.
	 * Update its place in the group list accordingly.
	 * Add any groups with changed propensity to the change list.
	 * @param r The updated reaction.
	 */
	private void updateGroups(Reaction r)
	{
		// Identify this reaction's group.
		PropensityGroup originalGroup = r.getGroup();		
		// Determine whether r remains in the same group or is reassigned.
		if(originalGroup == null || !originalGroup.revisePropensity(r)) {
			// r must be assigned to another group, or dropped.
			double propensity = r.getPropensity();
			if(propensity >= minimumPropensity) {
				int groupId = (int) logBase2(propensity / minimumPropensity);
				// Make sure the group exists.
				if(groups.size() <= groupId) {
					// This can be avoided if we have an accurate maximumPropensity.
					addNewGroup(groupId);
				}
				PropensityGroup revisedGroup = groups.get(groupId);
				revisedGroup.addReaction(r);
				// Update the propensity tree for the new group.
				int leafNode = propensityTreeLeafOffset + groupId;
				propensityTree[leafNode] = revisedGroup.getGroupPropensitySum();
				// Add the parent to the change list, as it needs to be updated.
				changeList[parentNode(leafNode)] = true;
			}
			else {
				// Consider it unreactable.
				// NB: It could be put in group 0, but with a low probability
				// of selection.
				r.setGroup(null);
			}
		}

		if(originalGroup != null) {
			// Update the propensity tree because originalGroup's propensity has changed.
			int leafNode = propensityTreeLeafOffset + originalGroup.getId();
			propensityTree[leafNode] = originalGroup.getGroupPropensitySum();
			// Add the parent to the change list, as it needs to be updated.
			changeList[parentNode(leafNode)] = true;
		}
	}
	
	/**
	 * Add a new group.
	 * This is only used while the model is running and is the result of failing
	 * to create sufficient groups before the model was run.
	 * This method can be avoided if pmax can be established reliably
	 * before the model is run.
	 * @param groupId The ID of the new group
	 */
	private void addNewGroup(int groupId)
	{
		for(int groupNumber = groups.size(); groupNumber <= groupId; groupNumber++) {
			double boundary = Math.pow(2.0, groupNumber + 1) * minimumPropensity;
			groups.add(new PropensityGroup(groupNumber, boundary));						
		}
		// If the number of groups has gone over a power-of-2 boundary then
		// the propensity tree must be completely rebuilt :(
		int depth = (int) Math.ceil(logBase2(groups.size()));
		int nodes = (int) (Math.pow(2.0, depth + 1) - 1);
		if(nodes > propensityTree.length) {
			// Rebuild the tree as it is not large enough.
			System.err.println("Tree expansion necessary from " + 
							   propensityTree.length + " to " + nodes);
			buildPropensityTree();
            // Resize the change list.
            int listSize = (int) Math.pow(2.0,
                                    (int) Math.ceil(logBase2(groups.size())));
            boolean[] oldChangeList = changeList;
            changeList = new boolean[listSize];
            System.arraycopy(oldChangeList, 0,
                             changeList, 0, oldChangeList.length);
		}		
	}
	
	/**
	 * Build the tree of cumulative propensity values.
	 * The sum for all groups will be at the root of the tree.
	 */
	private void buildPropensityTree()
	{
		// Calculate how much space is required for the tree, which is a complete
		// binary tree, storing the group propensity values in its leaves.
		int numGroups = groups.size();
		int depth = (int) Math.ceil(logBase2(numGroups));
		int nodes = (int) (Math.pow(2.0, depth + 1) - 1);
		int leaves = (int) Math.pow(2.0, depth);
		
		this.propensityTreeLeafOffset = nodes - leaves;
		this.propensityTree = new double[nodes];
		
		// Fill in the leaf values of the tree.
		int leafNode = propensityTreeLeafOffset;
		for(PropensityGroup group : groups) {
			propensityTree[leafNode] = group.getGroupPropensitySum();
			leafNode++;
		}
		
		// Pull the leaf values up to the root, which will hold the
		// cumulative propensity for all groups.
		// NB: An alternative would be to call changePropensityNodeValue in
		// the loop above each time a leaf's value is set, but that is
		// a relatively wasteful approach with a large tree.
		calculatePropensityChildren(0);
	}
	
	/**
	 * Show the propensity tree.
	 */
	private void showPropensityTree()
	{
		int level = 0;
		int nodesAtLevel = (int) Math.pow(2.0, level);
		for(int node = 0; node < propensityTree.length; node++) {
			if(node== nodesAtLevel-1) {
				System.out.println();
				level++;
				nodesAtLevel = (int) Math.pow(2.0, level);
			}
			System.out.print(node + " : " + propensityTree[node] + " ");
		}
		System.out.println();
	}
	
	/**
	 * Sum the child values of node to give the value for node.
	 * @param node The node to be computed.
	 * @return The computed value of the node.
	 */
	private double calculatePropensityChildren(int node)
	{
		int leftChild = 2 * node + 1;
		if(leftChild < propensityTree.length) {
			return propensityTree[node] = 
				calculatePropensityChildren(leftChild) + calculatePropensityChildren(leftChild + 1);
		}
		else {
			// Leaf.
			return propensityTree[node];
		}
	}
	
	/**
	 * Recalculate the value at the given node and percolate the change
	 * up to the root.
	 * @param node The node whose value must be recalculated.
	 */
	private int percolateChangedNodeValue(int node)
	{
		int leftChild = 2 * node + 1;
		propensityTree[node] = propensityTree[leftChild] + propensityTree[leftChild + 1];
		if(node > 0) {
			// Trickle the change up to the root.
			return 1 + percolateChangedNodeValue(parentNode(node));
		}
		else {
			return 1;
		}
	}
	
	/**
	 * Build the dependency graph for the reactions.
	 */
	private void buildDependencyGraph() 
	{
		Map<Reaction, Set<Reaction>> mapDependencyGraph =
			new HashMap<Reaction, Set<Reaction>>(reactions.size());
		// Map of molecules to the list of reactions depending on them.
		Map<Molecule, List<Reaction>> moleculeDependencyList = 
			new HashMap<Molecule, List<Reaction>>(molecules.size());
		
		// Build a map of which reactions depend on which molecules.
		for(Reaction target : reactions) {
			Set<Molecule> dependsOn = target.getDependsOn();
			for(Molecule m : dependsOn) {
				List<Reaction> dependency = moleculeDependencyList.get(m);
				if(dependency == null) {
					dependency = new ArrayList<Reaction>();
					moleculeDependencyList.put(m, dependency);
				}
				dependency.add(target);
			}
		}

		// For each reaction, identify which molecules it affects
		// and which reactions depend upon those molecules.
		for(Reaction source : reactions) {
			// Create the set for affected reactions.
			Set<Reaction> affectedReactions = new HashSet<Reaction>();
			mapDependencyGraph.put(source, affectedReactions);
			
			// Add the self dependency.
			affectedReactions.add(source);
			
			// Identify the reactions affected.
			Set<Molecule> affects = source.getAffects();
			for(Molecule m : affects) {
				List<Reaction> dependency = moleculeDependencyList.get(m);
				if(dependency != null) {
					// This reaction affects all the dependent reactions.
					affectedReactions.addAll(dependency);
				}
			}
		}		
		
		// Now flatten the map to a 2D array.
		dependencyGraph = new Reaction[mapDependencyGraph.size()][];
		for(Reaction r : reactions) {
			int id = r.getId();
			Set<Reaction> affectedReactions = mapDependencyGraph.get(r);
			dependencyGraph[id] = new Reaction[affectedReactions.size()];
			affectedReactions.toArray(dependencyGraph[id]);
			mapDependencyGraph.put(r, null);
		}		
	}
	
	/**
	 * Show the dependency graph for the reactions.
	 */
	private void showDependencyGraph()
	{
		System.out.println("Dependencies");
		for(Reaction source : reactions) {
			System.out.println(source);
			Reaction[] affectedReactions = dependencyGraph[source.getId()];
			for(Reaction target : affectedReactions) {
				System.out.println("    " + target);
			}			
		}		
		System.out.println();
	}
	
	/**
	 * Return the base 2 log of value
	 * @param value
	 * @return the base 2 log of value
	 */
	private double logBase2(double value)
	{
		return Math.log(value) / LOG2;
	}
	
	/**
	 * Return the parent node of the given child.
	 * @param childNode
	 * @return The child's parent node.
	 */
	private int parentNode(int childNode)
	{
		return (childNode - 1) / 2;
	}
	
	/**
	 * Internal consistency check.
	 */
	private void checkTotalPropensity()
	{
		double sum = 0.0;
		for(PropensityGroup group : groups) {
			sum += group.getGroupPropensitySum();
		}
		assert Math.abs(propensityTree[0] - sum) < 1e-5 :
			"Tree: Should be: " + propensityTree[0] + " is: " + sum;
	}

}
